/* $Id: bnm2dsp.c,v 1.4 2003/12/22 13:03:49 damien Exp $ */

/*-
 * Copyright (c) 2003, Damien Bergamini <damien.bergamini@free.fr>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
#if defined(__RCSID)
__RCSID("$Id: bnm2dsp.c,v 1.4 2003/12/22 13:03:49 damien Exp $");
#else
#ifndef lint
static char rcsid[] = "$Id: bnm2dsp.c,v 1.4 2003/12/22 13:03:49 damien Exp $";
#endif /* not lint */
#endif

#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>

#define MAXDSPSIZE	(1024 * 1024)
#define MAXBNMLINELEN	(1 + 1 + 2 + 8 + (2 * 256))

static unsigned char *dsp;
static unsigned int start = -1;
static unsigned int dsplen = 0;
static char line[MAXBNMLINELEN];

static void usage(void);
static void read_bnm(char *);
static void write_dsp(char *);

//#define DIAGNOSTIC 1

int
main(int argc, char **argv)
{
	if (argc < 3)
		usage();

	if ((dsp = malloc(MAXDSPSIZE)) == NULL)
		err(EX_OSERR, NULL);

	argv++; /* skip progname */
	argc--;
	for (; argc > 1; argc--, argv++)
		read_bnm(*argv);

	write_dsp(*argv);

	return EX_OK;
}

static void
usage(void)
{
	extern char *__progname;

	(void)fprintf(stderr, "usage: %s first.bnm [second.bnm ...] dst.dsp\n", 
	    __progname);
	exit(EX_USAGE);
}

static unsigned int 
readhex(unsigned char *p, int len)
{
	unsigned int res = 0;

	while (len-- > 0) {
		res <<= 4;

		if (*p >= '0' && *p <= '9')
			res |= *p - '0';
		else if (*p >= 'a' && *p <= 'f')
			res |= *p - 'a' + 10;
		else if (*p >= 'A' && *p <= 'F')
			res |= *p - 'A' + 10;
		else
			errx(EX_DATAERR, "Unexpected character");

		p++;
	}

	return res;
}

static void
read_bnm(char *bnmname)
{
	FILE *fp;
	unsigned char *p;
	unsigned int i, len, ofs;
	unsigned int lineno;

	if ((fp = fopen(bnmname, "r")) == NULL)
		err(EX_NOINPUT, "%s", bnmname);

	lineno = 0;
	while (fgets(line, sizeof line, fp) == line) {
		lineno++;
		p = line;

		if (*p != 's' && *p != 'S')
			errx(EX_DATAERR, "Bad S-Record in file %s line %u", 
			    bnmname, lineno);

		p++; /* skip s or S */

		switch (*p) {
		case '2':
			p++; /* skip 2 */
			len = readhex(p, 2) - 4;
			p += 2;
			ofs = readhex(p, 6);
			p += 6;
			break;

		case '3':
			p++; /* skip 3 */
			len = readhex(p, 2) - 4;
			p += 2;
			ofs = readhex(p, 8);
			p += 8;
			break;

		case '9':
			(void)fclose(fp);
			return;

		default:
			errx(EX_DATAERR, "Unknown S-Record type in file %s "
			    "line %u", bnmname, lineno);
		}

		if (start == -1)
			start = ofs;

		/* check that dsplen do not overflow */
		if (ofs + len > dsplen)
			dsplen = ofs + len;

		for (i = 0; i < len; i++, p += 2)
			*(dsp + ofs + i) = (unsigned char)readhex(p, 2);
	}

	(void)fclose(fp);
}

/* macros to handle unaligned data in the DSP code */
#define DSP_GETB(p)  ((p)[0])
#define DSP_GETW(p)  ((p)[0] << 8  | (p)[1])
#define DSP_GETDW(p) ((p)[0] << 24 | (p)[1] << 16 | (p)[2] << 8 | (p)[3])
static void
write_dsp(char *dspname)
{
	FILE *fp;
	unsigned char *p, *pp, *boot = NULL;
	unsigned char  pagecount, blockcount;
	unsigned short blockaddr, blocksize, blockextsize, blockpadextsize;
	unsigned int  *taboffset, mainpageoffset, pageoffset;
	int i, j;

	if ((fp = fopen(dspname, "w")) == NULL)
		err(EX_CANTCREAT, "%s", dspname);

	p = dsp + start;

	mainpageoffset = DSP_GETDW(p);
	p += 4;
	pagecount = DSP_GETB(p);
	p += 1;

#if DIAGNOSTIC
	(void)printf("DSP Informations\n");
	(void)printf("Page Count: %u\n", pagecount);
#endif

	pagecount++; /* account for the main page */
	(void)fwrite(&pagecount, 1, 1, fp);

	if ((taboffset = calloc(pagecount, 4)) == NULL)
		err(EX_OSERR, NULL);

	(void)fwrite(taboffset, pagecount * 4, 1, fp);

	for (i = 1; i < pagecount; i++) {

		pageoffset = DSP_GETDW(p);
		p += 4;

#if DIAGNOSTIC
		(void)printf("Page %u at offset 0x%08x\n", i, pageoffset);
#endif

		if (pageoffset == 0)
			continue;

		pp = dsp + pageoffset;

		blockcount = DSP_GETB(pp);
		pp += 1;

#if DIAGNOSTIC
		(void)printf("\tBlock Count %u\n", blockcount);
#endif

		taboffset[i] = ftell(fp);
		(void)fwrite(&blockcount, 1, 1, fp);

		for (j = 0; j < blockcount; j++) {

			blockaddr = DSP_GETW(pp);
			pp += 2;
			blocksize = DSP_GETW(pp);
			pp += 2;
			blockextsize = DSP_GETW(pp);
			pp += 2;

#if DIAGNOSTIC
			(void)printf("\tBlock %u Address 0x%04x Size %u "
			    "ExtSize %u\n", j, blockaddr, blocksize, 
			    blockextsize);
#endif

			(void)fwrite(&blockaddr, 2, 1, fp);

			if (blockaddr & 0x4000) {

				(void)fwrite(&blockextsize, 2, 1, fp);
				(void)fwrite(pp, blockextsize, 1, fp);
				pp += blockextsize;

			} else {

				blockpadextsize = (blockextsize / 63) * 64 + 
				    (blockextsize % 63);
				(void)fwrite(&blockpadextsize, 2, 1, fp);

				while (blockextsize > 63) /* >= 64 */ {
					(void)fwrite(pp, 63, 1, fp);
					(void)fputc(0xff, fp); /* pad byte */

					pp += 63;
					blockextsize -= 63;
				}
				(void)fwrite(pp, blockextsize, 1, fp);
				pp += blockextsize;
			}
		}
	}

#if DIAGNOSTIC
	(void)printf("Main Page at offset 0x%08x\n", mainpageoffset);
#endif

	pp = dsp + mainpageoffset;

	blockcount = DSP_GETB(pp);
	pp += 1;

#if DIAGNOSTIC
	(void)printf("\tBlock Count %u\n", blockcount);
#endif

	taboffset[0] = ftell(fp);

	blockcount++; /* add a fictive block for bootstrap code */
	(void)fwrite(&blockcount, 1, 1, fp);

	for (j = 0; j < blockcount - 1; j++) {
		pp = dsp + mainpageoffset + 1 + j * 0xc00;

		blockaddr = DSP_GETW(pp);
		pp += 2;
		blocksize = DSP_GETW(pp);
		pp += 2;
		blockextsize = DSP_GETW(pp),
		pp += 2;

#if DIAGNOSTIC
		(void)printf("\tBlock %u Address %x Size %u Extended Size %u\n",
		    j, blockaddr, blocksize, blockextsize);
#endif

		if (blockaddr & 0x4000) {
			blocksize *= 2;
			blockextsize = blocksize;

			(void)fwrite(&blockaddr, 2, 1, fp);
			(void)fwrite(&blockextsize, 2, 1, fp);
			(void)fwrite(pp, blockextsize, 1, fp);

		} else {
			blocksize *= 3;
			blockextsize = blocksize;

			if (blockaddr == 0) {
				boot = pp;
				pp += 3;

				blockaddr += 1;
				blocksize -= 3;
				blockextsize = blocksize;
			}

			blockpadextsize = (blockextsize / 63) * 64 + 
			    (blockextsize % 63);
			(void)fwrite(&blockaddr, 2, 1, fp);
			(void)fwrite(&blockpadextsize, 2, 1, fp);

			while (blockextsize > 63) /* >= 64 */ {
				(void)fwrite(pp, 63, 1, fp);
				(void)fputc(0xff, fp); /* pad byte */

				pp += 63;
				blockextsize -= 63;
			}
			(void)fwrite(pp, blockextsize, 1, fp);
		}
	}

	/* finally, write the fictive block */
	if (boot == NULL)
		errx(EX_DATAERR, "No boot block found!");

	blockaddr = 0;
	blocksize = 3;
	blockextsize = blocksize;

	(void)fwrite(&blockaddr, 2, 1, fp);
	(void)fwrite(&blockextsize, 2, 1, fp);
	(void)fwrite(boot, blockextsize, 1, fp);

	/* write back the up-to-date page offsets */
	(void)fseek(fp, 1, SEEK_SET);
	(void)fwrite(taboffset, pagecount * 4, 1, fp);

	(void)fclose(fp);
}
