/*****************************************************************************/
/* Copyright 2002 Christian Casteyde (casteyde.christian@free.fr)            */
/*                                                                           */
/* eaglectrl is free software; you can redistribute it                       */
/* and/or modify it under the terms of the GNU General Public License as     */
/* published by the Free Software Foundation; either version 2 of the        */
/* License, or (at your option) any later version.                           */
/*                                                                           */
/* "ADI USB ADSL Driver for Linux" is distributed in the hope that it will   */
/* be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with "ADI USB ADSL Driver for Linux"; if not, write to the Free     */
/* Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,            */
/* MA  02111-1307  USA                                                       */
/*                                                                           */
/* Written by Fred. Ros (sl33p3r@free.fr) based on the adictrl utils by      */
/* Christian Casteyde (casteyde.christian@free.fr)                           */ 
/*                                                                           */
/*****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <dirent.h>
#include <fcntl.h>
#include <getopt.h>
#include <signal.h>
#include <assert.h>


/* ----------------------------------- Macros ----------------------------------- */

#define REVISION "-1.0"

#define EAGLECTRLVERSION EAGLEUSBVERSION REVISION


#define SKIP_BLANKS(p)                                          \
    while ((*(p) !='\0') &&                                     \
           ((*(p) == ' ') || (*(p) == '\t') || (*(p) == '\n'))) \
    {                                                           \
        p++;                                                    \
    }

#define SKIP_NON_BLANKS(p)                                      \
    while ((*(p) !='\0') &&                                     \
           ((*(p) != ' ') && (*(p) != '\t') && (*(p) != '\n'))) \
    {                                                           \
        p++;                                                    \
    }


#define REMOVE_CHAR(p,c)                        \
{                                               \
    char *__ptr= (p);                           \
                                                \
    while ((*__ptr != '\0') && (*__ptr != c))   \
    {                                           \
        __ptr ++;                               \
    }                                           \
    if ( *__ptr == c )                          \
    {                                           \
        *__ptr = '\0';                          \
    }                                           \
}

#define NB_OPTIONS (sizeof(eu_options_t)/sizeof(eu_opt_t))

#define MAX_CMVS_COUNT   100


struct ueagle_cmvs {
        u_int32_t address;
        u_int16_t offset;
        u_int32_t data;
} __attribute__ ((packed));

#define MAKESA(a, b, c, d)                                              \
        (((c) & 0xff) << 24 |                                           \
         ((d) & 0xff) << 16 |                                           \
         ((a) & 0xff) << 8  |                                           \
         ((b) & 0xff))

static char * outfile = "CMV.bin";

/**
 * Read the CMVs file.
 *
 * filename: path to the CMVs file
 * drv_cmvs: pointer to CMV array.
 * nb_cmvs: number of read CMVs. When the function is called it should hold the
 *          max. number of CMVs the array can hold.
 *
 * return: EXIT_FAILURE / EXIT_SUCCESS
 */
static int read_cmv_from_file (
                               const char *filename,
                               struct ueagle_cmvs *drv_cmvs,
                               unsigned int *nb_cmvs
                              ) 
{
    struct ueagle_cmvs    *pcmvs   = drv_cmvs;
    char          line[256];
    char         *ptr;
    FILE         *fp      = NULL;
    unsigned int max_cmvs;
    int           retcode = EXIT_SUCCESS;
    
    pcmvs = drv_cmvs;
    max_cmvs = *nb_cmvs;
    *nb_cmvs = 0;
    
    
    fp = fopen (filename,"rt");
    if (fp == NULL) 
    {
        fprintf (stderr,"CMV file %s does not exists\n",filename);
        retcode = EXIT_FAILURE;
        goto byebye;
    }

    /*
     * Now treat file
     */
    pcmvs = drv_cmvs;    
    while (feof (fp) == 0 ) 
    {
        if ( fgets (line, 256, fp) == NULL ) 
        {
            continue;
        }

        if ( (*nb_cmvs) >= (max_cmvs-1) ) 
        {
            fprintf (stderr,"ERR: maximum number of CMVs (%u) reached.\n",
                     max_cmvs-1);
            retcode = EXIT_FAILURE;
            goto byebye;
        }

        
        /* Skip spaces at beginning */
        ptr = &line[0];
        
        REMOVE_CHAR(ptr,'\n');
        SKIP_BLANKS (ptr);

        switch (*ptr) 
        {
            case '\0':          /* Line empty */
            case '\r':          /* Line empty */
                continue;
                break;

            case '#':
            case ';':           /* Comments */
                continue;
                break;
        }
        /* Next 2 chars should be CW */
        if ( strncasecmp (ptr,"CW",2) != 0 ) 
        {
            fprintf (stderr,"[ERR] CMV file line \"%s\" not treated.\n",line);
            retcode = EXIT_FAILURE;
            goto byebye;
        }

        /* Advance ptr */
        ptr += 2;
        
        
        SKIP_BLANKS (ptr);
        if ( *ptr == '\0' ) 
        {
            fprintf (stderr,"[ERR] CMV file line too short (%s)\n",line);
            retcode = EXIT_FAILURE;
            goto byebye;
        }
        
        
        /* Next 4 chars are either cntl or optn */
/*         if ( ( strncasecmp (ptr,"OPTN",4) != 0 ) && */
/*              ( strncasecmp (ptr,"CNTL",4) != 0 ) )  */
/*         { */
/*             fprintf (stderr,"[ERR] CMV file line \"%s\" not treated (OPTN/CNTL)\n",line); */
/*             retcode = EXIT_FAILURE; */
/*             goto byebye; */
/*         } */

/*
 * FIXME: Add a check on the type of message. Currently in last version
 * furnished by ADI there are: optn, mask, diag, flag, cntl
 * 
 */
        
	pcmvs->address = MAKESA(toupper(ptr[0]), toupper(ptr[1]), toupper(ptr[2]), toupper(ptr[3]));

        /* Advance ptr */
        ptr += 4;

        SKIP_BLANKS (ptr);
        if ( *ptr == '\0' ) 
        {
            fprintf (stderr,"[ERR] CMV file line too short (%s)\n",line);
            retcode = EXIT_FAILURE;
            goto byebye;
        }

        /* Next is the offset */
        pcmvs->offset = atoi(ptr);

        /* Advance */
        SKIP_NON_BLANKS (ptr);
        
        SKIP_BLANKS (ptr);
        if ( *ptr == '\0' ) 
        {
            fprintf (stderr,"[ERR] CMV file line too short (%s)\n",line);
            retcode = EXIT_FAILURE;
            goto byebye;
        }

        /* And Data */
        
        pcmvs->data = (unsigned int) strtoul (ptr,NULL,16);
        if (errno == ERANGE) 
        {
            fprintf (stderr,"Value of option overflows (%s)\n", line);
            retcode = EXIT_FAILURE;
            goto byebye;
        }

        pcmvs ++;
        (*nb_cmvs) ++;        
    }    

  byebye:
    if ( fp ) 
    {
        fclose (fp);
    }
    
    return(retcode);
}

static int send_cmvs ( const char *file ) 
{
    int                   fd      = -1;
    int                   retcode = EXIT_SUCCESS;
    struct ueagle_cmvs             drv_cmvs[MAX_CMVS_COUNT];
    unsigned int          nb_cmvs = 0;
    
        
    /*
     * Set max. number of CMVS in the furnished array
     */
    nb_cmvs = MAX_CMVS_COUNT;
    
    retcode = read_cmv_from_file ( file, drv_cmvs, &nb_cmvs );
    if ( retcode != EXIT_SUCCESS ) 
    {
        fprintf (stderr,"Failed while reading CMV file.\n");
        goto byebye;
    }
    
    fd = creat(outfile, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);


    if ( fd == -1 ) 
    {
        perror ("Unable to send CMVs to driver");
        retcode = EXIT_FAILURE;
        goto byebye;
    }

    write(fd, &nb_cmvs, 1);
    write(fd, drv_cmvs, nb_cmvs*sizeof(struct ueagle_cmvs));
    close(fd);

    printf ("CMVs builded.\n");
    
  byebye:
    if ( fd != -1 ) 
    {
        close (fd);
    }
    
    return (retcode);
    
}


int main (int argc, char *argv[] ) 
{
	if (argc < 2) {
		printf("%s cmv_file\n",argv[0]);
		exit (1);
	}
	if (argc == 3)
		outfile = argv[2];

            return send_cmvs ( argv[1]);
}
